/*
** timeout.c - Generic timeout code
**
** Copyright (c) 1997-2000 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "plib/config.h"

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>

#include "plib/safeio.h"
#include "plib/safestr.h"
#include "plib/threads.h"
#include "plib/aalloc.h"

#include "plib/timeout.h"

#ifdef HAVE_THREADS


static struct timeout_cb
{
    pthread_mutex_t mtx;
    pthread_cond_t cv;
    pthread_t tid;
    
    int running;

    TIMEOUT *top;
} tcb;

pthread_once_t timeout_once = PTHREAD_ONCE_INIT;


static void *
timeout_thread(void *misc)
{
    struct timeout_cb *tcb = (struct timeout_cb *) misc;
    TIMEOUT *tp;
    time_t now;

    
    pthread_mutex_lock(&tcb->mtx);
    while (tcb->running)
    {
	time(&now);

	tp = tcb->top;
	while (tp && now >= tp->when)
	{
	    tcb->top = tp->next;
	    pthread_mutex_unlock(&tcb->mtx);
	    
	    if (tp->fun)
		tp->fun(tp->arg);
	    
	    pthread_mutex_lock(&tcb->mtx);
	    tp = tcb->top;
	}
	
	if (tp == NULL)
	    pthread_cond_wait(&tcb->cv, &tcb->mtx);
	else
	{
	    struct timespec when;

	    when.tv_sec = tp->when;
	    when.tv_nsec = 0;
	    
	    pthread_cond_timedwait(&tcb->cv, &tcb->mtx, &when);
	}
    }
    pthread_mutex_unlock(&tcb->mtx);
    
    return NULL;
}


static void
timeout_init(void)
{
    int ecode;
    
    pthread_mutex_init(&tcb.mtx, NULL);
    pthread_cond_init(&tcb.cv, NULL);

    tcb.top = NULL;
    tcb.running = 1;
    
    ecode = pthread_create(&tcb.tid, NULL, timeout_thread, &tcb);
    if (ecode)
    {
	syslog(LOG_ERR, "pthread_create(timeout_thread) failed: %s",
	       strerror(ecode));
	s_abort();
    }
}


TIMEOUT *
timeout_create(int timeout,
	       void (*fun)(void *arg),
	       void *arg)
{
    TIMEOUT *tp;
    

    pthread_once(&timeout_once, timeout_init);
    
    A_NEW(tp);

    tp->next = NULL;
    tp->when = 0;

    tp->fun = fun;
    tp->arg = arg;

    if (timeout_reset(tp, timeout) != 0)
    {
	timeout_destroy(tp);
	return NULL;
    }
    
    return tp;
}


int
timeout_reset(TIMEOUT *tp,
	      int timeout)
{
    TIMEOUT **prev, *cur;
    
    
    pthread_mutex_lock(&tcb.mtx);

    if (timeout > 0)
	tp->when = time(NULL)+timeout;
    else
	tp->when = 0;

    /* Locate it in the timeout list */
    prev = &tcb.top;
    cur = tcb.top;
    while (cur != NULL && cur != tp)
    {
	prev = &cur->next;
	cur = cur->next;
    }

    /* Remove it from the list, if found */
    if (cur == tp)
    {
	*prev = cur->next;
	tp->next = NULL;
    }

    if (timeout > 0)
    {
	/* Reinsert it at the new position */
	prev = &tcb.top;
	cur = tcb.top;
	while (cur != NULL && cur->when < tp->when)
	{
	    prev = &cur->next;
	    cur = cur->next;
	}
	*prev = tp;
	tp->next = cur;

    }
    
    pthread_mutex_unlock(&tcb.mtx);
    pthread_cond_signal(&tcb.cv);
    
    return 0;
}


#else /* No threads */

#include <signal.h>

static TIMEOUT *top = NULL;


static RETSIGTYPE
sigalarm_handler(int sig)
{
    time_t now;
    int timeout;
    

    
    s_signal(SIGALRM, sigalarm_handler);

    time(&now);
    
    while (top != NULL && top->when <= now)
    {
	top->fun(top->arg);
	top = top->next;
    }
    
    if (top)
    {
	timeout = top->when-now;
	alarm(timeout);
    }
}


TIMEOUT *
timeout_create(int timeout,
	      void (*fun)(void *arg),
	      void *arg)
{
    TIMEOUT *tp;
    

    A_NEW(tp);
    
    tp->next = NULL;
    tp->when = 0;
    tp->fun = fun;
    tp->arg = arg;

    if (timeout_reset(tp, timeout) != 0)
    {
	timeout_destroy(tp);
	return NULL;
    }
    
    return tp;
}


int
timeout_reset(TIMEOUT *tp,
	      int timeout)
{
    TIMEOUT **prev, *cur;
    time_t now;
    

    time(&now);

    if (timeout > 0)
	tp->when = now+timeout;
    else
	tp->when = 0;

    alarm(0);
    
    /* Locate it in the timeout list */
    prev = &top;
    cur = top;
    while (cur != NULL && cur != tp)
    {
	prev = &cur->next;
	cur = cur->next;
    }

    /* Remove it from the list, if found */
    if (cur == tp)
    {
	*prev = cur->next;
	tp->next = NULL;
    }

    if (timeout > 0)
    {
	/* Reinsert it at the new position */
	prev = &top;
	cur = top;
	while (cur != NULL && cur->when < tp->when)
	{
	    prev = &cur->next;
	    cur = cur->next;
	}
	*prev = tp;
	tp->next = cur;
    }

    if (top)
    {
	time(&now);
	
	timeout = top->when-now;

	if (timeout <= 0)
	    sigalarm_handler(SIGALRM);
	else
	    alarm(timeout);
    }
    
    return 0;
}

#endif


int
timeout_destroy(TIMEOUT *tp)
{
    if (tp == NULL)
	return -1;

    timeout_reset(tp, 0);
    a_free(tp);
    
    return 0;
}


